import json
import logging
from pyspark.sql import SparkSession
from delta import configure_spark_with_delta_pip
from ._utils import is_databricks

_logger = logging.getLogger(__name__)


def get_spark_context(project: str, config: dict = None):
    if is_databricks():
        _logger.debug("Getting databricks spark context")
        from databricks.sdk.runtime import spark

        return spark
    else:
        _logger.debug("Getting local spark context")

        if config is None:
            config = {
                "spark.master": "local",
                # "spark.jars.packages": io.delta:delta-core_2.12:2.1.1
                # "spark.sql.extensions": io.delta.sql.DeltaSparkSessionExtension
                "spark.sql.catalog.spark_catalog": "org.apache.spark.sql.delta.catalog.DeltaCatalog",
                "spark.databricks.delta.merge.repartitionBeforeWrite.enabled": True,
                "spark.sql.catalog.spark_catalog": "org.apache.spark.sql.delta.catalog.DeltaCatalog",
                "spark.sql.catalogImplementation": "hive",
            }

        msg = json.dumps(config, indent=4, default=str)
        _logger.debug(msg)

        builder = SparkSession.builder

        for k, v in config.items():
            builder = builder.config(k, v)

        builder.appName(project)
        spark = configure_spark_with_delta_pip(builder).getOrCreate()
        return spark



from ._deltalake import DeltaLake
from ._read import Read
from ._dataset import DataSet, Table
from ._factory import factory as dataset_factory


__all__ = ["DeltaLake", "Read", "dataset_factory", "DataSet", "Table"]



import logging
from pydantic import Field, PrivateAttr
from .._utils import JinjaVariables, render_jinja, is_databricks
from typing import Any, Dict, Union, List
from .._timeslice import Timeslice
import os
from .._stage_type import StageType
from ._dataset import DataSet, Table
from ..datallake import create_database, create_table

# try:
#     from databricks.sdk.runtime import *  # noqa F403
# except ModuleNotFoundError:
#     pass


class DeltaLakeTable(Table):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

    depends_on: List[str] = Field(default=[])
    delta_properties: Dict[str, str] = Field(default=None)
    delta_constraints: Dict[str, str] = Field(default=None)
    partition_by: List[str] = Field(default=None)
    z_order_by: List[str] = Field(default=None)
    create_table: bool = Field(default=True)
    managed: bool = Field(default=False)


class DeltaLake(DataSet, DeltaLakeTable):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self._logger = logging.getLogger(self.__class__.__name__)
        self._render()
        self.create_delta_table()

    @classmethod
    def in_allowed_stages(cls, stage: StageType):
        return stage in (stage.raw, stage.base, stage.curated)

    _logger: Any = PrivateAttr(default=None)
    _replacements: Dict[JinjaVariables, str] = PrivateAttr(default=None)
    options: Union[dict, None] = Field(default=None)
    timeslice: Timeslice = Field(...)
    location: str = Field(default=None)
    checkpoint_location: str = Field(default=None)
    stage: StageType = Field(...)
    managed: bool = Field(default=False)
    create_table: bool = Field(default=True)

    def _render(self):
        self._replacements = {
            JinjaVariables.TABLE: self.table,
            JinjaVariables.DATABASE: self.database,
            JinjaVariables.CONTAINER: self.container,
            JinjaVariables.CHECKPOINT: self.checkpoint,
        }

        self.root = render_jinja(self.root, self._replacements)
        self.path = render_jinja(self.path, self._replacements)
        self.database = render_jinja(self.database, self._replacements)
        self.table = render_jinja(self.table, self._replacements)
        if self.options:
            for option, value in self.options.items():
                self.options[option] = render_jinja(value, self._replacements)

        self.location = os.path.join(self.root, self.path)
        if not is_databricks():
            self.location = f"{self.config_path}/../data{self.location}"
            self.location = os.path.abspath(self.location)

    def create_delta_table(self):
        database_table = f"`{self.database}`.`{self.table}`"
        self._logger.info(f"Creating delta lake table {database_table}")
        create_database(self.database)

        if self.managed:
            create_table(
                database=self.database,
                table=self.table,
                delta_properties=self.delta_properties,
            )
        else:
            create_table(
                database=self.database,
                table=self.table,
                delta_properties=self.delta_properties,
                path=self.location,
            )



import logging
from pydantic import BaseModel, Field, PrivateAttr
from .._utils import JinjaVariables
from typing import Any, Dict, Union, List
from .._timeslice import Timeslice
from .._stage_type import StageType
from .dataset_type import TableType


class Table(BaseModel):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

    stage: StageType = Field(...)
    database: str = Field(...)
    table: str = Field(...)
    id: Union[str, List[str]] = Field(default=[])
    custom_properties: Dict[str, Any] = Field(default=None)
    table_type: TableType = Field(...)

    def _render(self):
        pass


class DataSet(BaseModel):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self._logger = logging.getLogger(self.__class__.__name__)

    _logger: Any = PrivateAttr(default=None)
    _replacements: Dict[JinjaVariables, str] = PrivateAttr(default=None)
    container: str = Field(...)
    root: str = Field(...)
    path: str = Field(default=None)
    options: dict = Field(...)
    timeslice: Timeslice = Field(...)
    checkpoint: str = Field(default=None)
    config_path: str = Field(...)

    def _render(self):
        pass



import json
from pyspark.sql import DataFrame
import logging
from pyspark.sql.types import StructType, StructField
import jinja2
from typing import List
from ._spark_context import get_spark_context

_logger = logging.getLogger(__name__)


def get_partition_predicate(partition_values: dict):
    predicates = []
    for k, v in partition_values.items():
        v = [str(val) for val in v]
        p = f"`{k}` in ({','.join(v)})"
        predicates.append(p)

    predicate = " and ".join(predicates)

    return predicate


def table_exists(database: str, table: str):
    spark = get_spark_context(project="test_project")
    table_exists = (
        spark.sql(f"SHOW TABLES in {database};")
        .where(f"tableName='{table}' AND !isTemporary")
        .count()
        == 1
    )
    return table_exists


def create_table(
    database: str,
    table: str,
    path: str = None,
    delta_properties: List[str] = None,
    sql: str = None,
):
    _logger.debug(f"Creating table if not exists {database}.{table} at {path}")

    if not sql:
        sql = f"CREATE TABLE IF NOT EXISTS `{database}`.`{table}`"

    # add in the delta properties if there are any
    sql_properties = ""
    if delta_properties:
        sql_properties = [
            f"{k.lower()} = {v.lower()}" for k, v in delta_properties.items()
        ]
        sql_properties = ", ".join(sql_properties)
        sql_properties = f"TBLPROPERTIES({sql_properties})"

    sql_path = ""
    if path:
        sql_path = f"USING DELTA LOCATION '{path}';"

        sql = f"{sql}\n{sql_path}\n{sql_properties}"

    spark = get_spark_context(project="test_project")
    _logger.debug(f"{sql}")
    spark.sql(sql)

    return sql


def create_database(database: str):
    _logger.debug(f"Creating database if not exists `{database}`")
    sql = f"CREATE DATABASE IF NOT EXISTS {database}"
    spark = get_spark_context(project="test_project")
    _logger.debug(sql)
    spark.sql(sql)
    return sql


def alter_table_drop_constraint(database: str, table: str, name: str):
    return f"ALTER TABLE `{database}`.`{table}` DROP CONSTRAINT {name};"


def alter_table_add_constraint(database: str, table: str, name: str, constraint: str):
    return f"ALTER TABLE `{database}`.`{table}` ADD CONSTRAINT {name} CHECK ({constraint});"


def alter_table_set_tblproperties(database: str, table: str, properties: str):
    return f"ALTER TABLE `{database}`.`{table}` SET TBLPROPERTIES ({properties});"


def get_table_properties(database: str, table: str):
    _logger.debug(f"getting existing table properties for table {database}.{table}")

    spark = get_spark_context(project="test_project")

    df: DataFrame = spark.sql(f"SHOW TBLPROPERTIES `{database}`.`{table}`").collect()
    tbl_properties = {
        c.asDict()["key"]: c.asDict()["value"]
        for c in df
        if not c["key"].startswith("delta.constraints")
    }
    tbl_constraints = {
        c.asDict()["key"].split(".")[-1]: c.asDict()["value"]
        for c in df
        if c["key"].startswith("delta.constraints")
    }

    properties = {
        f"{database}.{table}": {
            "constraints": tbl_constraints,
            "properties": tbl_properties,
        }
    }
    msg = json.dumps(properties, indent=4, default=str)
    _logger.debug(msg)
    return properties


def optimize(database: str, table: str, partition_values: dict, zorder_by: list = []):
    sql = f"OPTIMIZE `{database}`.`{table}`"

    if partition_values:
        predicate = get_partition_predicate(partition_values)
        predicate = f" WHERE {predicate}"
        sql = f"{sql}{predicate}"

    if zorder_by:
        sql_zorderby = ",".join([f"`{z}`" for z in zorder_by])
        sql = f"{sql} ZORDER BY ({sql_zorderby})"

    spark = get_spark_context(project="test_project")
    _logger.debug(f"optimizing table {database}.{table}\n{sql}")
    spark.sql(sql)


def get_table_details(database: str, table: str):
    _logger.debug(
        f"getting existing table details and partitions for table {database}.{table}"
    )
    spark = get_spark_context(project="test_project")

    df: DataFrame = spark.sql(
        f"DESCRIBE TABLE EXTENDED `{database}`.`{table}`"
    ).collect()

    # get the details into a dictionary
    details = {c.asDict()["col_name"]: c.asDict()["data_type"] for c in df}

    # pull out the columns
    columns = {}
    ordinal = 0
    for k, v in details.items():
        if k and v:
            columns[k] = {}
            columns[k]["ordinal"] = ordinal
            columns[k]["type"] = v
            ordinal = +1
        else:
            break

    # pull out the columns
    partitions = [v for k, v in details.items() if k.startswith("Part ")]

    details = {
        f"{database}.{table}": {
            "columns": columns,
            "partitions": partitions,
            "name": details.get("Name"),
            "location": details.get("Location"),
            "provider": details.get("Provider"),
            "owner": details.get("Owner"),
        }
    }

    msg = json.dumps(details, indent=4, default=str)
    _logger.debug(msg)
    return details


def create_column_ddl(field: StructField):
    nullable = "" if field.nullable else "NOT NULL"
    comment = f"COMMENT {field.metadata}" if field.metadata else ""
    field_type = field.dataType.typeName()
    field_name = f"`{field.name}`"

    return f"\t{field_name} {field_type} {nullable} {comment}"


def create_table_dll(
    schema: StructType,
    partition_fields: list = [],
    format: str = "DELTA",
    always_identity_column: str = None,
):
    field_ddl = [create_column_ddl(f) for f in schema.fields]
    if always_identity_column:
        always_identity_column = (
            f"\t`{always_identity_column}` GENERATED ALWAYS AS IDENTITY"
        )
        field_ddl = [always_identity_column] + field_ddl

    field_ddl = ",\n".join(field_ddl)

    template_partition = jinja2.Template("PARTITIONED BY ({{partition_fields}})")
    template_ddl = jinja2.Template(
        """CREATE TABLE {{database_name}}.{{table_name}}
(
{{field_ddl}}
)
USING {{format}} LOCATION '{{path}}'
{{partition_ddl}}""",
        undefined=jinja2.DebugUndefined,
    )

    if partition_fields:
        partition_fields = [f"`{p}`" for p in partition_fields]
        partition_fields = ",".join(partition_fields)
        partition_ddl: str = template_partition.render(
            partition_fields=partition_fields
        )
    else:
        partition_ddl = ""

    replace = {"field_ddl": field_ddl, "partition_ddl": partition_ddl, "format": format}

    table_ddl = template_ddl.render(replace)
    table_ddl = f"{table_ddl};"

    return table_ddl



import os
from .dataset import DataSet
from ._timeslice import Timeslice
from ._tables import Tables, _INDEX_WILDCARD
from ._stage_type import StageType
from .dataset import dataset_factory
from ._utils import abs_config_path, load_yaml, get_config_path, check_version
from ._logging_config import configure_logging
from ._project import Project


class Config:
    _TABLES = "tables"
    _CONFIG_PATH = "config_path"

    def __init__(self, project: str, pipeline: str, config_path: str = None):
        configure_logging(project)
        self.project = self._load_project(project, config_path)
        self.pipeline = self._load_pipeline(pipeline)
        self.tables = self._load_tables()

    def _load_project(self, project: str, config_path: str):
        self.config_path = get_config_path(project, config_path)
        project_file_path = os.path.join(self.config_path, f"{project}.yaml")
        project_config = load_yaml(project_file_path)
        check_version(project_config)
        project_config["config_path"] = self.config_path
        project = Project(**project_config)
        return project

    def _load_pipeline(self, pipeline: str):
        pipeline_file = f"{pipeline}.yaml"

        config_file_path = os.path.join(self.project.pipelines, pipeline_file)
        pipeline = load_yaml(config_file_path)
        check_version(pipeline)

        # add the configuration path into the confif dictionart
        # so that it gets passed to table config when created
        pipeline[self._CONFIG_PATH] = self.project.pipelines
        return pipeline

    def _load_tables(self):
        tables_path = self.pipeline[self._TABLES]
        tables_path = abs_config_path(
            self.project.pipelines, self.pipeline[self._TABLES]
        )

        data = load_yaml(tables_path)
        check_version(data)
        self.pipeline[self._TABLES] = data

        tables = Tables(
            table_data=self.pipeline[self._TABLES],
            config_path=self.pipeline[self._CONFIG_PATH],
        )
        return tables

    def get_table_mapping(
        self,
        timeslice: Timeslice,
        stage: StageType,
        table: str = _INDEX_WILDCARD,
        database: str = _INDEX_WILDCARD,
    ):
        table_mapping = self.tables.get_table_mapping(
            stage=stage, table=table, database=database
        )

        table_mapping.source = dataset_factory.get_data_set(
            self.pipeline, table_mapping.source, timeslice
        )
        table_mapping.destination = dataset_factory.get_data_set(
            self.pipeline, table_mapping.destination, timeslice
        )

        return table_mapping

    def set_checkpoint(
        self,
        source: DataSet,
        destination: DataSet,
        checkpoint_name: str = None,
    ):
        if not checkpoint_name:
            checkpoint_name = f"{source.database}.{source.table}-{destination.database}.{destination.table}"

        source.checkpoint = checkpoint_name
        source._render()
        destination.checkpoint = checkpoint_name
        destination.options["checkpointLocation"] = destination.checkpoint_location
        destination._render()
